import numpy as np
import scipy as sp

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

import os
from igor2 import binarywave as bw


def load_ibw(file, ss=False):
    '''
    Load the ibw file as an IBWData object.

    Input:
        file     - String: path to the ibw file
        ss         - Boolean: if True then the ibw file will be treated as domain switching
                 spectroscopy file.
    Output:
        IBWData object:
        self.z          - Numpy array: 2D numpy array containing topography channel. 
                         Default is the "Height" channel.
        self.size       - float: Map size in the unit of meter
        self.mode        - String: Imaging mode. Currently support "AC Mode", "Contact Mode", "PFM Mode"
                         "SS Mode" and "DART Mode".
        self.header     - Dict: All the setup information.
        self.channels   - list: List of channel names
        self.data       - Numpy array: An array of all the saved image data in this ibw file in
                         the same order as self.channels
    Examples:
        
    '''
    
    # Return an IBWData object
    return IBWData(file, ss=ss)

class IBWData(object):
    '''
    Data structure for AR IBW maps.

    Attributes:
        self.z          - Numpy array: 2D numpy array containing topography channel. 
                         Default is the "Height" channel.
        self.size       - float: Map size in the unit of meter
        self.mode        - String: Imaging mode. Currently support "AC Mode", "Contact Mode", "PFM Mode"
                         "Spec" and "DART Mode".
        self.header     - Dict: All the setup information.
        self.channels   - list: List of channel names
        self.data       - Numpy array: An array of all the saved image data in this ibw file in
                         the same order as self.channels
        
    Methods:
        None
    '''
    def __init__(self, path, ss=False):
        super(IBWData, self).__init__()

        self._load_ibw(path)

        # Spectroscopy files:
        if ss == True:
            self.mode = "Spec"
            try:
                self._load_ss()
            except IndexError:
                pass
        elif "ARDoIVCurve" in self.header:
            self.mode = "Spec"
            try:
                self._load_ss()
            except IndexError:
                pass
        # Image files:
        else:
            try:
                self.size = self.header['ScanSize']
                self.mode = self.header['ImagingMode']
                z_index = self.channels.index('Height')
                self.z = self.data[z_index]
                
                try:
                    # Separate DART mode from general PFM mode
                    if self.mode == "PFM Mode":
                        if len(self.channels) > 4:
                            self.mode = "DART Mode"
                            self.channels = ['Height', 'Amplitude1', 'Amplitude2', 'Phase1', 'Phase2', 'Frequency']
                        else:
                            self.channels = ['Height', 'Amplitude', 'Deflection', 'Phase']
                except IndexError:
                    pass
            except KeyError:
                pass

    def _load_ibw(self, path):

        t = bw.load(path)
        wave = t.get('wave')

        # Decode the notes section to parse the header
        if isinstance(wave['note'], bytes):
            try:
                parsed_string = wave['note'].decode('utf-8').split('\r')
            except:
                parsed_string = wave['note'].decode('ISO-8859-1').split('\r')

        # Load the header
        self.header = {}

        for item in parsed_string:
            try:
                key, value = item.split(':', 1)
                value = value.strip()  # Remove leading/trailing whitespace
            except ValueError:
                continue  # For items that do not split correctly

            # Determine the data type of the value and convert
            if '.' in value or 'e' in value:  # Floating point check
                try:
                    self.header[key] = float(value)
                except ValueError:
                    self.header[key] = value
            elif value.lstrip('-').isdigit():  # Integer check
                self.header[key] = int(value)
            else:
                self.header[key] = value

        # Load the data
        # Transpose the data matrix
        self.data = wave['wData'].T
        # data = wave['wData']

        # Load the channel names
        self.channels = [self.header.get(f'Channel{i+1}DataType', 'Unknown') for i in range(np.shape(self.data)[0])]

    def _load_ss(self, nan=True):

        if nan is True:
            bias_raw= self.data[-1]
            index_not_nan = np.where(~np.isnan(bias_raw))

            bias = bias_raw[index_not_nan]
            amp, phase1, phase2, freq = self.data[2][index_not_nan], self.data[3][index_not_nan], \
                        self.data[4][index_not_nan], self.data[5][index_not_nan]
        else:
            bias = self.data[-1]
            amp = self.data[2]
            phase1 = self.data[3]
            phase2 = self.data[4]
            freq = self.data[5]

        #Correcting first without offset to calculate the drive parameters
        phase1 = self._correct_phase_wrapping(phase1, offset_correction=False)
        phase2 = self._correct_phase_wrapping(phase2, offset_correction=False)

        df = self.header['DFRTFrequencyWidth']

        a_dr, ph_dr, q = self._calc_drive_params(amp, amp, phase1/180*np.pi, phase2/180*np.pi, freq, df)

        phase1 = self._correct_phase_wrapping(phase1)
        phase2 = self._correct_phase_wrapping(phase2)
        ph_dr  = self._correct_phase_wrapping(ph_dr/np.pi*180)

        # Let's count how many times the bias has changed (on->off, off->on)
        index_bp = np.where(np.diff(bias) != 0)[0] + 1
        index_bp = np.concatenate([[0], index_bp])

        # Output array length
        length = len(index_bp) // 2

        bias_on, bias_off = np.zeros(length), np.zeros(length)

        phase1_on,phase1_off = np.zeros(length), np.zeros(length)
        phase2_on, phase2_off  = np.zeros(length), np.zeros(length)
        amp_on, amp_off = np.zeros(length), np.zeros(length)
        freq_on, freq_off = np.zeros(length), np.zeros(length)

        amp_dr_on, amp_dr_off = np.zeros(length), np.zeros(length)
        phase_dr_on, phase_dr_off = np.zeros(length), np.zeros(length)
        q_on, q_off = np.zeros(length), np.zeros(length)

        for i in range(length * 2-1):
            if i % 2 == 0: # bias off
                phase1_off[i//2] = np.mean(phase1[index_bp[i]:index_bp[i+1]])
                phase2_off[i//2] = np.mean(phase2[index_bp[i]:index_bp[i+1]])
                amp_off[i//2] = np.mean(amp[index_bp[i]:index_bp[i+1]])
                freq_off[i//2] = np.mean(freq[index_bp[i]:index_bp[i+1]])
                bias_off[i//2] = np.mean(bias[index_bp[i]:index_bp[i+1]])

                phase_dr_off[i // 2] = np.mean(ph_dr[index_bp[i]:index_bp[i + 1]])
                q_off[i // 2] = np.mean(q[index_bp[i]:index_bp[i + 1]])
                amp_dr_off[i // 2] = np.mean(a_dr[index_bp[i]:index_bp[i + 1]])

            else:
                bias_on[i//2] = np.mean(bias[index_bp[i]:index_bp[i+1]])
                phase1_on[i//2] = np.mean(phase1[index_bp[i]:index_bp[i+1]])
                phase2_on[i//2] = np.mean(phase2[index_bp[i]:index_bp[i+1]])
                amp_on[i//2] = np.mean(amp[index_bp[i]:index_bp[i+1]])
                freq_on[i//2] = np.mean(freq[index_bp[i]:index_bp[i+1]])

                phase_dr_on[i // 2] = np.mean(ph_dr[index_bp[i]:index_bp[i + 1]])
                q_on[i // 2] = np.mean(q[index_bp[i]:index_bp[i + 1]])
                amp_dr_on[i // 2] = np.mean(a_dr[index_bp[i]:index_bp[i + 1]])

        self.bias = bias_on[1:]
        self.phase1_on = phase1_on[1:]
        self.phase1_off = phase1_off[1:]
        self.phase2_on = phase2_on[1:]
        self.phase2_off = phase2_off[1:]
        self.freq_on = freq_on[1:]
        self.freq_off = freq_off[1:]
        self.amp_on = amp_on[1:]
        self.amp_off = amp_off[1:]
        self.x_on = amp_on[1:] * np.cos(phase1_on[1:]/180*np.pi)
        self.x_off = amp_off[1:] * np.cos(phase1_off[1:] / 180 * np.pi)

        self.amp_dr_on = amp_dr_on[1:]
        self.amp_dr_off = amp_dr_off[1:]
        self.x_dr_on = amp_dr_on[1:] * np.cos(phase_dr_on[1:] / 180 * np.pi)
        self.x_dr_off = amp_dr_off[1:] * np.cos(phase_dr_off[1:] / 180 * np.pi)
        self.phase_dr_on = phase_dr_on[1:]
        self.phase_dr_off = phase_dr_off[1:]
        self.q_on = q_on[1:]
        self.q_off = q_off[1:]

        # return bias[1:], amp_off[1:], phase1_off[1:], phase2_off[1:]

    def _correct_phase_wrapping(self, ph, lower=-90, upper=270, offset_correction=True):
        '''
        Correct the phase wrapping in Jupiter.
        
        Input:
            Ph     - Array: array of phase values
            lower - float: lower bound of phase limit in your instrument
            upper - float: upper bound of phase limit in your instrument
        Output:
            ph_shift - Array: phase with wrapping corrected
        '''
        # Use the phase value measured at last pixel as the offset in the lock-in
        if offset_correction:
            ph_shift = ph - ph[-1]
        else:
            ph_shift = ph

        index_upper = np.where(ph_shift > upper)
        index_lower = np.where(ph_shift < lower)
        ph_shift[index_upper] -= 360
        ph_shift[index_lower] += 360

        return ph_shift

    @staticmethod
    def _calc_drive_params(_a1, _a2, _ph1, _ph2, _fc, _df):
        '''
        Calculate real Dart parameters from the observables.

        Input:
            _a1  - amplitude 1
            _a2  - amplitude 2
            _ph1 - phase 1
            _ph2 - phase 2
            _fc  - resonance frequency
            _df  - difference between freq 2 and freq 1
        Output:
            _a_drive  - drive amplitude
            _ph_drive - resonance phase
            _q        - resonanse quality factor
        '''

        epsilon = 1e-10  # a small adding for calculation stability
        _dph = _ph2 - _ph1
        _f1 = _fc - _df / 2
        _f2 = _fc + _df / 2

        _om = _f1 * _a1 / (_f2 * _a2)
        _fi = np.tan(_dph)

        _x1 = -(1 - np.sign(_fi) * np.sqrt(1 + np.square(_fi)) / _om) / (_fi + epsilon)
        _x2 = (1 - np.sign(_fi) * np.sqrt(1 + np.square(_fi)) * _om) / (_fi + epsilon)

        _q = np.sqrt(_f1 * _f2 * (_f2 * _x1 - _f1 * _x2) * (_f1 * _x1 - _f2 * _x2)) / (np.square(_f2) - np.square(_f1))
        _q[_q > 1000] = 1000
        _a_drive = _a1 * np.sqrt((_fc**2 - _f1**2)**2 +(_fc * _f1 / _q)**2) / np.square(_fc)
        _ph_drive = _ph1 - np.arctan(_fc * _f1 / (_q * (np.square(_fc) - np.square(_f1))))

        return _a_drive, _ph_drive, _q


# Function to read and plot all the ibw files

def display_ibw_folder(folder, mode=None, key=['Height'], cmaps=None, paras=None, save=None, **kwargs):
    '''
    Display all the ibw files with specified modes in a given folder.

    Input:
        folder  - Required: path to the folder to be explored
        mode    - Optional: if not given, all available modes will be displayed
                            can be 'AC Mode', 'PFM Mode', 'Contact Mode', 'Spec' and 'DART Mode'
                            To be added: 'SKPM Mode'
        key     - Optional: list of channels to be displayed
        cmaps     - Optional: list of color maps that will be used for each channels in key
        paras   - Optional: to be added for visualizing imaging parameters like surface voltage
        save     - Optional: if None, no image will be saved. If not None, each image will be 
                            saved as fileName + save
        **kwarg - Optional: Additional keyword arguments are sent to imshow().

    Output:
        ibw_files   -list: ibw file names in the same order as they are displayed
        data        -list: SciFiReader object of each ibw file displayed

    Example use:
        ibw_files, data = display_ibw(folder, mode='AC Mode', key=['Height', 'ZSensor'])
    '''
    file_names = os.listdir(folder)
    file_names = sorted(file_names)

    file_displayed = []
    out = []

    if key is not None:
        if not isinstance(key, list):
            key = list(key)

    display_index = 0
    for index, file in enumerate(file_names):
        if file.endswith('.ibw'):
            fname = os.path.join(folder, file)

            try:
                t = load_ibw(fname)

                if save is not None:
                    save = save + ' ' + file.split('.')[0]

                if t.mode != 'Spec': # skip the spectrum ibw files
                    if mode == None:
                        display_ibw(t, key=key, display_index=display_index, cmaps=cmaps, save=save, **kwargs)
                        file_displayed.append(fname)
                        out.append(t)
                        display_index += 1
                    elif mode == t.mode:
                        display_ibw(t, key=key, display_index=display_index, cmaps=cmaps, save=save, **kwargs)
                        file_displayed.append(fname)
                        out.append(t)
                        display_index += 1
                else:
                    pass
            except TypeError:
                pass

    return file_displayed, out

def display_ibw(file, key=None, titles=None, display_index=None, cmaps=None, save=None, **kwargs):
    '''
    Display a single ibw with specified by the file path.

    Input:
        file    - Required: path to the file to be displayed or the loaded ibw object returned by load_ibw()
        key     - Optional: list of channels to be displayed
        titles     - Optional: list of titles corresponding to the key
        display_index - Optional: index provided by display_ibw_folder() function
        cmaps     - Optional: list of color maps that will be used for each channels in key
        save     - Optional: if None, no image will be saved. If not None, each image will be 
                            saved as fileName + save
        **kwarg - Optional: Additional keyword arguments are sent to imshow().

    Output:
        ibw_files   -list: ibw file names in the same order as they are displayed
        data        -list: SciFiReader object of each ibw file displayed

    Example use:
        ibw_files, data = display_ibw(folder, mode='AC Mode', key=['Height', 'ZSensor'])
    '''

    try:
        if type(file) is str:
            t = load_ibw(file)
        else:
            t = file

        if key is not None:
            if not isinstance(key, list):
                key = list(key)
        else:
            key = t.channels

        if titles is not None:
            if not isinstance(titles, list):
                titles = list(titles)

        if cmaps is not None:
            if not isinstance(cmaps, list):
                cmaps = list(cmaps)

        if t.mode != 'Spec': # skip the spectrum ibw files
            indices = find_channel(obj=t, key=key)
            if len(indices) == 1: # Only one channel will be displayed
                plt.figure(figsize=[4,4])
                to_plot = t.data[indices[0]]
                if cmaps is None:
                    im = plt.imshow(to_plot, extent=[0, t.size*1e6, 0, t.size*1e6], **kwargs)
                else:
                    im = plt.imshow(to_plot, extent=[0, t.size*1e6, 0, t.size*1e6], cmap=cmaps[0], **kwargs)
                if display_index is None:
                    title = "{}: {}".format(t.mode, key[0]) if not titles else titles
                else:
                    title = "{}: {}-{}".format(display_index, t.mode, key[0]) if not titles else titles

                plt.title(title)
                divider = make_axes_locatable(plt.gca())
                cax = divider.append_axes("right", size="5%", pad=0.05)
                plt.colorbar(im, cax=cax)
                plt.tight_layout()
                if save is not None:
                    plt.savefig('{}.png'.format(save), dpi=400, bbox_inches='tight', pad_inches=0.1)
            else:
                n_cols = len(indices)
                fig,ax=plt.subplots(1, n_cols, figsize=[n_cols*3+1, 3])
                for i in range(len(indices)):
                    to_plot = t.data[indices[i]]
                    if cmaps is None:
                        im = ax[i].imshow(to_plot, extent=[0, t.size*1e6, 0, t.size*1e6], **kwargs)
                    else:
                        im = ax[i].imshow(to_plot, extent=[0, t.size*1e6, 0, t.size*1e6], cmap=cmaps[i], **kwargs)
                    divider = make_axes_locatable(ax[i])
                    cax = divider.append_axes("right", size="5%", pad=0.05)
                    fig.colorbar(im, cax=cax)
                    if titles is None:
                        if not i:
                            ax[i].set_title("{}: {}-{}".format(display_index, t.mode, t.channels[indices[i]]))
                        else:
                            ax[i].set_title("{}".format(t.channels[indices[i]]))
                    else:
                        ax[i].set_title(titles[i])
                    plt.tight_layout()
                    if save is not None:
                        plt.savefig('{}.png'.format(save), dpi=400, bbox_inches='tight', pad_inches=0.1)
        else:
            pass
    except TypeError:
        pass

def find_channel(obj, key):
    if key is None:
        return np.arange(len(obj.channels))
    else:
        index = []
        channels = obj.channels

        for item in key:
            if item in channels:
                index.append(channels.index(item))
        return index

def plane_subtract(A, pts=None):
    '''
    Remove a plane fitting from a map. If pts are given, the plane will be
    fitted based on pts. Otherwise, the three corner points will used to fit 
    for the plane.

    Inputs:
        A        - Required : A 2D numpy array.
        pts     - Optional : coordinates in pixels for three points: [pt1, pt2, pt3]
    Returns:
        A_out   - Map after a plane fit is removed
    '''

    H, W = np.shape(A)
    x = np.arange(W)
    y = np.arange(H)
    X, Y = np.meshgrid(x, y)
    
    if pts == None:
        p0 = [W-1,0]
        p1 = [0,H-1]
        p2 = [W-1,H-1]
    else:
        p0, p1, p2 = pts
    
    Ax, Ay, Az = *p0, A[int(p0[1]), int(p0[0])]
    Bx, By, Bz = *p1, A[int(p1[1]), int(p1[0])]
    Cx, Cy, Cz = *p2, A[int(p2[1]), int(p2[0])]
    
    a = (By-Ay)*(Cz-Az) - (Cy-Ay)*(Bz-Az)
    b = (Bz-Az)*(Cx-Ax) - (Cz-Az)*(Bx-Ax)
    c = (Bx-Ax)*(Cy-Ay) - (Cx-Ax)*(By-Ay)
    d = -(a*Ax+b*Ay+c*Az)
    z = -(a*X+b*Y+d) / c
    return A - z

def lineSubtract(data, n=1, maskon=False, thres=4, M=4, normalize=True, colSubtract=False):
    '''
    Remove a polynomial background from the data line-by-line, with
    the option to skip pixels within certain distance away from
    impurities.  If the data is 3D (eg. 3ds) this does a 2D background
    subtract on each layer independently.  Input is a numpy array.

    Inputs:
        data    -   Required : A 1D, 2D or 3D numpy array.
        n       -   Optional : Degree of polynomial to subtract from each line.
                               (default : 1).
        maskon  -   Optional : Boolean flag to determine if the impurty areas are excluded.
        thres   -   Optional : Float number specifying the threshold to determine
                               if a pixel is impurity or bad pixels. Any pixels with intensity greater
                               than thres*std will be identified as bad points.
        M       -   Optional : Integer number specifying the box size where all pixels will be excluded
                               from poly fitting.
        normalize - Optional : Boolean flag to determine if the mean of a layer
                               is set to zero (True) or preserved (False).
                               (default : True)
        colSubtract - Optional : Boolean flag (False by default) to determine if polynomial background should also be subtracted column-wise

    Returns:
        subtractedData  -   Data after removing an n-degree polynomial
    '''
    # Polyfit for lineSubtraction excluding impurities.
    def filter_mask(data, thres, M, D):
        filtered = data.copy()
        if D == 1:
            temp = np.gradient(filtered)
            badPts = np.where(np.abs(temp-np.mean(temp))>thres*np.std(temp))
            for ix in badPts[0]:
                filtered[max(0, ix-M) : min(data.shape[0], ix+M+1)] = np.nan
            return filtered
        elif D == 2:
            temp = np.gradient(filtered)[1]
            badPts = np.where(np.abs(temp-np.mean(temp))>thres*np.std(temp))
            for ix, iy in zip(badPts[1], badPts[0]):
                filtered[max(0, iy-M) : min(data.shape[0], iy+M+1),
                                  max(0, ix-M) : min(data.shape[1], ix+M+1)] = np.nan
            return filtered

    def subtract_mask(data, n, thres, M, D):
        d = data.shape[0]
        x = np.linspace(0,data.shape[-1]-1,data.shape[-1])
        filtered = filter_mask(data, thres, M, D)
        output = data.copy()
        if D == 1:
            index = np.isfinite(filtered)
            try:
                popt = np.polyfit(x[index], data[index], n)
                output = data - np.polyval(popt, x)
            except TypeError:
                raise TypeError('Empty x-array encountered. Please use a larger thres value.')
            return output
        if D == 2:
            for i in range(d):
                index = np.isfinite(filtered[i])
                try:
                    popt = np.polyfit(x[index], data[i][index], n)
                    output[i] = data[i] - np.polyval(popt, x)
                except TypeError:
                    raise TypeError('Empty x-array encountered. Please use a larger thres value.')
            return output

    if maskon is not False:
        if len(data.shape) == 3:
            output = np.zeros_like(data)
            for ix, layer in enumerate(data):
                output[ix] = subtract_mask(layer, n, thres, M, 2)
            return output
        elif len(data.shape) == 2:
            return subtract_mask(data, n, thres, M, 2)
        elif len(data.shape) == 1:
            return subtract_mask(data, n, thres, M, 1)
        else:
            raise TypeError('Data must be 1D, 2D or 3D numpy array.')

    # The original lineSubtract code.
    def subtract_1D(data, n):
        x = np.linspace(0,1,len(data))
        popt = np.polyfit(x, data, n)
        return data - np.polyval(popt, x)
    def subtract_2D(data, n):
        if normalize:
            norm = 0
        else:
            norm = np.mean(data)
        output = np.zeros_like(data)
        for ix, line in enumerate(data):
            output[ix] = subtract_1D(line, n)
        if colSubtract:
            temp = np.zeros_like(data)
            for ix, line in enumerate(np.transpose(output)):
                temp[ix] = subtract_1D(line, n)
            output = np.transpose(temp)
        return output + norm

    if len(data.shape) == 3:
        output = np.zeros_like(data)
        for ix, layer in enumerate(data):
            output[ix] = subtract_2D(layer, n)
        return output
    elif len(data.shape) == 2:
        return subtract_2D(data, n)
    elif len(data.shape) == 1:
        return subtract_1D(data, n)
    else:
        raise TypeError('Data must be 1D, 2D or 3D numpy array.')
    

def fft(dataIn, window='None', output='absolute', zeroDC=False, beta=1.0,
        units='None'):
    '''
    Compute the fast Frouier transform of a data set with the option to add
    windowing.

    Inputs:
        dataIn    - Required : A 1D, 2D or 3D numpy array
        window  - Optional : String containing windowing function used to mask
                             data.  The options are: 'None' (or 'none'), 'bartlett',
                             'blackman', 'hamming', 'hanning' and 'kaiser'.
        output  - Optional : String containing desired form of output.  The
                             options are: 'absolute', 'real', 'imag', 'phase'
                             or 'complex'.
        zeroDC  - Optional : Boolean indicated if the centeral pixel of the
                                FFT will be set to zero.
        beta    - Optional : Float used to specify the kaiser window.  Only
                               used if window='kaiser'.
        units   - Optional : String containing desired units for the FFT.
                             Options: 'None', or 'amplitude' (in the future, I
                             might add "ASD" and "PSD".

    Returns:
        fftData - numpy array containing FFT of data

    Usage:
        fftData = fft(data, window='None', output='absolute', zeroDC=False,
                      beta=1.0)
    '''
    def ft2(data):
        ftData = np.fft.fft2(data)
        if zeroDC:
            ftData[0,0] = 0
        return np.fft.fftshift(ftData)

    outputFunctions = {'absolute':np.absolute, 'real':np.real,
                       'imag':np.imag, 'phase':np.angle, 'complex':(lambda x:x) }

    windowFunctions = {'None':(lambda x:np.ones(x)), 'none':(lambda x:np.ones(x)),
                       'bartlett':np.bartlett, 'blackman':np.blackman,
                       'hamming':np.hamming, 'hanning':np.hanning,
                       'kaiser':np.kaiser , 'sine':windows.cosine}

    outputFunction = outputFunctions[output]
    windowFunction = windowFunctions[window]

    data = dataIn.copy()
    if zeroDC:
        if len(data.shape) == 3:
            for ix, layer in enumerate(data):
                data[ix] -= np.mean(layer)
        else:
            data -= np.mean(data)

    if len(data.shape) != 1:
        if window == 'kaiser':
            wX = windowFunction(data.shape[-2], beta)[:,None]
            wY = windowFunction(data.shape[-1], beta)[None,:]
        else:
            wX = windowFunction(data.shape[-2])[:,None]
            wY = windowFunction(data.shape[-1])[None,:]
        W = wX * wY
        if len(data.shape) == 2:
            wData = data * W
            ftData = outputFunction(ft2(wData))
        elif len(data.shape) == 3:
            wTile = np.tile(W, (data.shape[0],1,1))
            wData = data * wTile
            if output == 'complex':
                ftData = np.zeros_like(data, dtype=np.complex)
            else:
                ftData = np.zeros_like(data)
            for ix, layer in enumerate(wData):
                ftData[ix] = outputFunction(ft2(layer))
        else:
            print('ERR: Input must be 1D, 2D or 3D numpy array')

    else:
        if window == 'kaiser':
            W = windowFunction(data.shape[0], beta)
        else:
            W = windowFunction(data.shape[0])
        wData = data * W
        ftD = np.fft.fft(wData)
        ftData = outputFunction(np.fft.fftshift(ftD))
    if units == 'amplitude':
        if len(data.shape) == 3:
            datashape = data[0].shape
        else:
            datashape = data.shape
        for size in datashape:
            ftData /= size
            if window == 'hanning':
                ftData *= 2
            elif window == 'None' or window == 'none':
                pass
            else:
                print('WARNING: The window function "%s" messes up the FT units' %
                        window)
    return ftData




from igor2 import binarywave as bw

import time
import os
from subprocess import Popen
import numpy as np
import matplotlib.pyplot as plt
from IPython.display import display, clear_output

import shutil
import subprocess

import types
import pickle

import aespm
from aespm.utils import shared

import platform

if platform.system() == 'Windows':
    _buffer_path = os.path.join(os.path.expanduser('~'), 'Documents', 'buffer')
    _command_buffer = os.path.join(_buffer_path, 'ToIgor.arcmd')
    _read_out_buffer = os.path.join(_buffer_path, 'readout.txt')
    _bash_buffer = os.path.join(_buffer_path, 'SendToIgor.bat')
    _path_txt = os.path.join(_buffer_path, 'path.txt')
    
    with open(_path_txt, 'r') as fopen:
        _exe_path = fopen.readline()
    
    # _exe_path = r"C:\AsylumResearch\v19\RealTime\Igor Pro Folder\Igor.exe"
else:
    _buffer_path = r"C:\Users\Asylum User\Documents\buffer"
    _command_buffer = r"C:\Users\Asylum User\Documents\buffer\ToIgor.arcmd"
    _read_out_buffer = r"C:\Users\Asylum User\Documents\buffer\readout.txt"
    _bash_buffer = r"C:\Users\Asylum User\Documents\buffer\SendToIgor.bat"
    _exe_path = r"C:\AsylumResearch\v19\RealTime\Igor Pro Folder\Igor.exe"

class Experiment(object):
    '''
    Experiment data structure for SPM workflows.

    Attributes:
        obj.param      - Dict: A collection of parameters can be updated with obj.update_param
        obj.folder     - String: Data saving folder on the local computer. Img/Spec data will be retrieved as 
                            the latest modified file in this folder.
        obj.temp_folder- String: Data exchange folder on the local computer. Commands and parameter reading buffer files
                            will be saved in this folder.
        obj.action_list- List: Names of custom functions. It will override default action list if doubled
        obj.log        - List: logged operations/parameter changes/reading
        obj.connection - SSH connection: return by utils.return_connection()
        obj.client     - SSH Client that can directly run commands on the local computer: return by utils.return_connection()
        
    Methods:
        obj.execute          - Execute a single operation, a wrapper of spm_control()
        obj.execute_sequence - Execute a list of operations in sequence
        obj.update_param     - update parameters in obj.param
        obj.add_func         - Add a custom function as the new method of obj
        obj.save_exp         - Save the whole experiment obj as .pkl file
        obj.log_operation    - log all the operations in obj.log
        obj.print_log        - print the logged operations nicely

    Usage:
        # Connection: [host, username, password]
        exp = Experiment(folder=your_folder, temp_folder=you_temp_folder, connection=connectio)
    '''
    def __init__(self, folder, temp_folder=r"C:\Users\Asylum User\Documents\AEtesting\data_exchange", connection=None):
        super(Experiment, self).__init__()

        # log of all the operations
        self.log = [] 
        # Initialize parameter dict 
        self.param = {} 
        # Initialize action list of custom functions 
        self.action_list = {}
        # This is the folder where your measurements are saved
        self.folder = folder

        # This is the folder where your temp data/command files are saved
        self.temp_folder = temp_folder
        
        # Remote control
        if connection is not None:
            host, username, password = connection
            local_info = aespm.utils.get_local_directory(host=host).split('$')
            global shared
            #shared = SharedInfo(host, local_info)--> future implementation
            shared.set_values((_buffer_path, _command_buffer, _read_out_buffer, _bash_buffer, _exe_path))
            shared.set_host(host)
            self.connection, self.client = aespm.utils.return_connection(host, username, password)

        else:
            self.connection, self.client = None, None

    def execute(self, action, value=None, wait=None, log=True, **kwargs):
        '''
        A wrapper of spm_control() function. 

        Input:
            action      - String: SPM instructions in hyper-language.
            value       - Int/Float/String: New value for the parameter to change.
            wait        - Float: sleep time after the action is finished.
            log         - Boolean: If true, this action will be logged in obj.log
            kwargs      - Keyword arguments for custom functions

        Return:
            N/A

        Examples:
            # Start a downward scan
            obj.execute(action='ScanDown')
            # Change scan rate to 1 Hz
            obj.execute(action='ScanRate', value=1)
        '''
        
        wait = 0.35 if wait is None else wait
        if wait <= 0.35:
            wait = 0.35

        # custom functions
        if action in self.action_list:
            if value == None:
                return getattr(self, action)()
            else:
                if type(value) != list and type(value) != np.ndarray:
                    value = [value]
                return getattr(self, action)(*value, **kwargs)
            time.sleep(wait)

        # default action list 
        else:
            if type(value) != list and type(value) != np.ndarray:
                value = [value]
            try:
                spm_control(action=action, value=value[0], wait=wait, connection=self.connection)
            except KeyError:
                print("Invalid action: not found in default list or custom list.")

        if log is True:
            self.log_operation(operation=[action, value, kwargs], dtype=0)

    def execute_sequence(self, operation, log=True, **kwargs):
        '''
        Execute multiple operations sequentially. 

        Input:
            operation - List: [[Action1, Value1, Wait_Time1], ..., [ActionN, ValueN, Wait_TimeN]]
                               if there are keyword args, using [[Action1, Value1, Wait_Time1, Dict1], ...]
            log       - Boolean: If true, all operations will be logged in obj.log
            kwargs    - Keyword arguments for custom functions

        Output:
            N/A

        Usage:
            obj.execute_sequence(operation=[['Setpoint', 0.1, None], ['DownScan', None, 1.5]])
        '''
        for item in operation:
            if item[0] in self.action_list:
                if len(item) > 3:
                    self.execute(action=item[0], value=item[1], wait=item[2], log=log, **item[-1])
                else:
                    self.execute(action=item[0], value=item[1], wait=item[2], log=log)
            else:
                self.execute(action=item[0], value=item[1], wait=item[2], log=log)

    def update_param(self, key, value, log=True):
        '''
        Update the value stored in obj.param

        Input:
            key     - List: keys to be modifies in obj.param
            value   - List: values to be entered in obj.param['key']
            log     - Boolean: If true, all operations will be logged in obj.log

        Output:
            N/A

        Usage:
            obj.update_param(key=['DriveAmplitude', 'Setpoint'], value=[0.1, 0.2])
        '''
        if type(key) is not list:
            key = [key]
        
        if len(key) == 1:
            self.param[key[0]] = value
            
        else:
            for i, ix in enumerate(key):
                self.param[ix] = value[i]
                if log is True:
                    self.log_operation(operation=[ix], dtype=1)
                    # self.log_operation(operation=[ix, value[i]], dtype=1)

    def log_operation(self, operation, dtype=0, **kwargs):
        '''
        Formats and logs an operation in hyper-language syntax in obj.log.
        Input:
            operation - List: [Action, Value, Wait_Time], the same as inputs used in spm_control()
            dtype      - Int: 0 - Operations/actions, 1 - Parameter update, 2 - Read data 
        Output:
            N/A

        Usage:
            obj.log(operation=['DownScan', None, 1.5]) # Start a downscan
            obj.log(operation=['Setpoint', 0.1, None]) # Change setpoint to 0.1 V
            obj.log(operation=[['DriveAmplitude', 'Setpoint'], [0.1, 0.2], type=1)
        '''
        to_log = {}
        to_log[operation[0]] = operation
        to_log['type'] = dtype
        self.log.append(to_log)

    def add_func(self, NewFunc, log=True):
        '''
        Add a custom function as the method to Experiment object.

        Input:
            NewFunc - Function: Custom function defined by user. This function has aceess 
                        to all the attributes and methods of obj
        Output:
            N/A 
        Usage:
            def measure(self, operation, key, value):
                self.update_param(key=key, value=value)
                self.execute(operation)
            obj.add_func(measure)
        '''
        # method_name = NewFunc.__name__
        # setattr(self, method_name, NewFunc)
        # getattr(self, method_name).__doc__ = NewFunc.__doc__

        # # add new function name to the custom action list
        # self.action_list.append(method_name)

        method_name = NewFunc.__name__
        
        # add new function name to the custom action list
        self.action_list[method_name] = method_name
        
        # Bind the function as a method of the instance
        bound_method = types.MethodType(NewFunc, self)
        
        # # Optionally, update the docstring (if necessary)
        # getattr(self, method_name).__doc__ = NewFunc.__doc__

        # Set the method to the instance
        setattr(self, method_name, bound_method)
        

        if log is True:
            self.log_operation(operation=[method_name], dtype=3)

    def save_exp(self, save_name):
        '''
        Save the whole experiment object as .pkl file.
        Input: 
            save_name - String: save name for the .pkl file
        Output:
            N/A
        Usage:
            obj.save_exp('my exp')
        '''

        with open('{}.pkl'.format(save_name), 'wb') as outp:
            pickle.dump(self, outp, pickle.HIGHEST_PROTOCOL)

    def print_log(self):
        '''
        Nicely plot the logged operations.
        '''
        for i, log_item in enumerate(self.log):
            if log_item['dtype'] == 0: # Normal operations
                print("Operation - {}: Action={}, Value={}, wait={}".format(i, 
                    log_item['operation'][0], log_item['operation'][1], log_item['operation'][2]))
            elif log_item['dtype'] == 1: # Normal operations
                print("Update - {}: Param={}, Value={}".format(i, 
                    log_item['operation'][0], log_item['operation'][1]))
            elif log_item['dtype'] == 2: # Read data
                print("Reading - {}: Data={}, Value={}".format(i, 
                    log_item['operation'][0], log_item['operation'][1]))
            elif log_item['dtype'] == 3: # Add custom function
                print("Add new function - {}: Name={}".format(i, 
                    log_item['operation'][0]))

            else:
                print("Unsupported operations in the log.")

def write_spm(commands, connection=None, wait=0.35):
    '''
    Control Jupiter by writing and executing Igor command in the ToIgor.arcmd file.

    Input:
        commands   - String: AR command strings to be executed by Igor.exe
        wait       - Float: sleep time after the writing process is done.
        connection - SSH connection: return by utils.return_connection()

    Return:
        N/A

    Usage:
        # Enable the x PISLoop
        write_spm('ARExecuteControl("StartPIDSLoop0","PIDSLoopPanel",0,"")')

        # Enable the x PISLoop by remote control
        connection = return_connection()
        write_spm('ARExecuteControl("StartPIDSLoop0","PIDSLoopPanel",0,"")', connection=connection)
    '''
    if connection==None:
        file = open(_command_buffer,"w",encoding = 'utf-8')
        file.writelines(commands)
        file.close()

        p = Popen(_bash_buffer)
        p.wait()
        time.sleep(wait)
    else:
        aespm.utils.write_to_remote_file(connection,
                             file_path = _command_buffer, data = commands)
        aespm.utils.main_exe_on_server(host=shared._host)
        time.sleep(wait)


def read_spm(key, commands=None, connection=None):
    '''
    Read Jupiter parameters by exporting them into readout.txt file.

    Input:
        key       - list: keys in AR to read the parameter
        commands  - String: user-defined command to read parameters
        connection - SSH connection: return by utils.return_connection()

    Returns:
        An array of parameter values specified by key.

    Usage:
        # Read out x and y LVDT sensitivity
        keys = ['XLVDTSens', 'YLVDTSens']
        xsens, ysens = read_spm(key=keys)
    '''
    if type(key) != list and type(key) != np.ndarray:
        key = [key]
    if connection == None:
        if commands == None:
            N = len(key)
            start = 'Make/N = ({})/O ReadOut // change N to number of parameters to read out\n'.format(N)
            command = ''
            for i in range(N):
                if key[i].startswith('PIDSLoop'): # == 'PIDSLoop.0.Setpoint' or key[i] == 'PIDSLoop.1.Setpoint':
                    command += 'ReadOut[{}] = td_ReadValue("{}")\n'.format(i, key[i])
                else:
                    command += 'ReadOut[{}] = GV("{}")\n'.format(i, key[i])
            end = r'Save/O/G/J ReadOut as "{}"'.format(_read_out_buffer.replace('\\', '\\\\'))
            file = open(_command_buffer,"w",encoding = 'utf-8')
            file.writelines(start+command+end)
            file.close()
            p = Popen(_bash_buffer)
            p.wait()
            return np.loadtxt(_read_out_buffer)

        else:
            file = open(_command_buffer,"w",encoding = 'utf-8')
            file.writelines(commands)
            file.close()

            p = Popen(_bash_buffer)
            p.wait()
            return np.loadtxt(_read_out_buffer)

    else:
        if commands == None:
            N = len(key)
            start = 'Make/N = ({})/O ReadOut // change N to number of parameters to read out\n'.format(N)
            command = ''
            for i in range(N):
                if key[i].startswith('PIDSLoop'): # == 'PIDSLoop.0.Setpoint' or key[i] == 'PIDSLoop.1.Setpoint':
                    command += 'ReadOut[{}] = td_ReadValue("{}")\n'.format(i, key[i])
                else:
                    command += 'ReadOut[{}] = GV("{}")\n'.format(i, key[i])
            end = r'Save/O/G/J ReadOut as "{}"'.format(_read_out_buffer.replace('\\', '\\\\'))
            commands = start+command+end
            aespm.utils.write_to_remote_file(connection,
                         file_path = _command_buffer, data = commands)# doubt as richard
            aespm.utils.main_exe_on_server(host=shared._host)

            s = aespm.utils.read_remote_file(connection, _read_out_buffer)

            return [float(k) for k in s.decode('utf-8').split('\r')[:-1]]

        else:
            aespm.utils.write_to_remote_file(connection,
                             file_path = _command_buffer, data = commands)# doubt as richard

            aespm.utils.main_exe_on_server(host=shared._host)
            s = aespm.utils.read_remote_file(connection, _read_out_buffer)

            return [float(k) for k in s.decode('utf-8').split('\r')[:-1]]
        
def spm_control(action, value=None, wait=0.35, connection=None):
    '''
    Control SPM with action. An action-command dict is used to convert the action/input to the actual AR commands.

    Input:
        action      - String: SPM instructions in hyper-language.
        value       - Int/Float/String: New value for the parameter to change.
        wait        - Float: sleep time after the action is finished.
        connection  - SSH connection: return by return_connection()
    Return:
        N/A

    Examples:
        # Start a downward scan
        spm_control(action='ScanDown')
        # Change scan rate to 1 Hz
        spm_control(action='ScanRate', value=1)
    '''
    # key: action, value[0]: input/button name, value[1]: window/panel name, 
    # value[-1]: 
    #       0-Button, 1-numeric, 2-string, 4-pure commands, 
    #       5-use value[-2] for numeric input, 6-use value[-2] for string input
    action_dict = {}
    key_list = [
        ['ScanRate', 'ScanRates', 'speed'], # 0 
        ['DownScan', 'ScanDown', 'Start', 'start'],
        ['UpScan', 'ScanUp'],
        ['DriveAmp', 'DriveAmplitude', 'DriveVoltage', 'DriveVolt', 'v_ac', 'V_ac'],
        ['IGain', 'IGains', 'IntegralGain', 'igain', 'I'], # 4
        ['DriveFreq', 'DriveFrequency', 'Freq', 'Frequency'],
        ['Setpoint', 'SetPoint'],
        ['StopScan', 'Stop', 'stop'],
        ['ClearForce', 'Clear', 'ClearMarker'],
        ['ThatsIt', 'Thats'], # 9
        ['GoThere', 'Gothere'],
        ['TipVoltage', 'V_tip', 'v_tip'],
        ['SurfaceVoltage', 'V_surface', 'v_surface', 'v_surf'],
        ['SingleF', 'SingleForce'],
        ['EnableStage', 'StageEnable', 'EnableMotor', 'MotorEnable'], # 14
        ['Approach', 'StartApproach'], 
        ['DisableStage', 'StageDisable', 'DisableMotor', 'MotorDisable'],
        ['ChangeName', 'FileName', 'BaseName'],
        ['GetTune', 'TuneData'],
        ['IVAmp', 'IV_Amp', 'iv_amp'], # 19
        ['IVFreq', 'IV_Freq', 'iv_freq'],
        ['IVDoIt', 'ivdoit', 'IVDoit'],
        ['FBHeight', 'FeedbackHeight', 'PID2=Height', 'FB_Height', 'FB_height'],
        ['StartFB2', 'Engage', 'StartLoop2', 'StartLoopZ'],
        ['IVAmpDART', 'IV_Amp_DART', 'iv_amp_DART'], # 24
        ['IVFreqDART', 'IV_Freq_DART', 'iv_freq_DART'],
        ['IVDoItDART', 'ivdoit_DART', 'IVDoit_DART'],
        ["OneTuneDART", "TuneDART", "DARTTune", "DART_Tune"],
        ['StartFB0', 'StartLoop0', 'StartLoopX'],
        ['StartFB1', 'StartLoop1', 'StartLoopY'], # 29
        ['PFM_Mode', 'Mode=PFM', 'PFM'],
        ['AC_Mode', 'Mode=AC', 'AC'],
        ['ScanSize', 'Scan Size', 'scan size', 'scan_size'],
        ['Points and Lines', 'Points', 'Pixels', 'points', 'pixels'],
        ['DARTFreq', 'DART Freq', 'CentralFreq'], #34
        ['DARTPhase1', 'DART Phase1'],
        ['DARTPhase2', 'DART Phase2'],
        ['DARTWidth', 'DART Width', 'FreqWidth'],
        ['CenterPhase', 'Center Phase', 'PhaseCenter'],
        ['Cycles', 'Cycle', 'NumCycles'], # 39
        ['DualFreq', 'DARTMode', 'DARTDualFreq'],
        ['DARTSweepWidth', 'SweepWidth', ],
        ['Capture', 'TakePhoto'],
        ['ACTune','AC Tune', 'Tune'],
        ['XOffset', 'X Offset', 'Offset_X'], # 44
        ['YOffset', 'Y Offset', 'Offset_Y'],
        ['ScanAngle', 'Scan Angle', 'Angle'],
        ['SetpointAmp', 'AC Setpoint', 'SetpointAC'],
        ['SetpointDefl', 'PFMSetpoint', 'SetpointPFM', 'SetpointContact'],
        ['BlueDriveAmp', 'BlueDriveAmplitude'], # 49
        ['FDTrigger', 'FDTriggerPoint'],
        ['ZeroPD', 'PDZero'],
        ['DARTTrigger', 'SSTrigger'],
        ['DARTAmp', 'DART v_ac'],
        ['SampleHeight', 'Sample Height'],
        
    ]

    value_list = [
        ['ScanRateSetVar_0', 'MasterPanel', 1], # 0
        ['DownScan_0', 'MasterPanel', 0],
        ['UpScan_0', 'MasterPanel', 0],
        ['DriveAmplitudeSetVar_0', 'MasterPanel', 1],
        ['IntegralGainSetVar_0', 'MasterPanel', 1], # 4
        ['DriveFrequencySetVar_0', 'MasterPanel', 1],
        ['td_WriteValue("Cypher.PIDSLoop.2.SetPoint", {})'.format(value), 4], #['td_WriteValue("Cypher.PIDSLoop.2.SetPoint", {})'.format(value), 4],
        ['StopScan_0', 'MasterPanel', 0],
        ['ClearForce_1', 'MasterPanel', 0],
        ['ForceSpotNumberSetVar_1', 'MasterPanel', 0], # 9 
        ['GoForce_1', 'MasterPanel', 0],
        ['TipVoltageSetVar_0', 'NapPanel', 1],
        ['SurfaceVoltageSetVar_0', 'NapPanel', 1],
        ['SingleForce_1', 'MasterPanel', 0],
        ["EnableStageCB_0","MasterMotorPanel", 1, 5], # 14
        ["MotorEngageButton_0","MasterMotorPanel",0], 
        ["EnableStageCB_0","MasterMotorPanel", 0, 5],
        ['ChangeName("{}")'.format(value), 4],
        ['GetTune()', 4],
        ["ARDoIVAmpSetVar_1","ARDoIVPanel", 1], # 19
        ["ARDoIVFrequencySetVar_1","ARDoIVPanel", 1],
        ["ARDoIVDoItButton_1","ARDoIVPanel", 0],
        ["DefaultPIDSLoop2","PIDSLoopPanel", "Height", 6],
        ["StartPIDSLoop2","PIDSLoopPanel", 0],
        ["ARDoIVAmpSetVar_1","DARTSpectroscopy", 1], # 24
        ["ARDoIVFrequencySetVar_1","DARTSpectroscopy", 1],
        ["SingleForce_2","DARTSpectroscopy", 0],
        ["DoTuneDFRT_3","DART", 0],
        ["StartPIDSLoop0","PIDSLoopPanel", 0],
        ["StartPIDSLoop1","PIDSLoopPanel", 0], # 29
        ["ImagingModePopup_0","MasterPanel","PFM Mode", 6],
        ["ImagingModePopup_0","MasterPanel","AC Mode", 6],
        ['ScanSizeSetVar_0', 'MasterPanel', 1],
        ['PointsLinesSetVar_0', 'MasterPanel', 1],
        ["DFRTFrequencyCenterSetvar","DART", 1], # 34
        ["PhaseOffsetSetVar_3","DART", 1],
        ["PhaseOffset1SetVar_3","DART", 1],
        ["DFRTFrequencyWidthSetvar","DART", 1],
        ["DoTuneCent_3", "DART", 1], 
        ["ARDoIVCyclesSetVar_1","DARTSpectroscopy", 1], # 39
        ["DualACModeBox_3", "DART", 1], 
        ["SweepWidthSetVar_3", "DART", 1], 
        ["ARVCaptureButton_0","ARVideoPanel", 0],
        ["DoTuneOnceButton","TuneGraph", 0],
        ['PV("XOffset", {})'.format(value), 4], # 44
        ['PV("YOffset", {})'.format(value), 4],
        ['PV("ScanAngle", {})'.format(value), 4],
        ['PV("AmplitudeSetpointVolts", {})'.format(value), 4],
        ['PV("DeflectionSetpointVolts", {})'.format(value), 4],
        ['bdDriveAmplitudeSetVar_0', 'MasterPanel', 1], # 49
        ['TriggerPointSetVar_1', 'MasterPanel', 1],
        ["ZeroLDPDButton_1","MasterMotorPanel",0], 
        ["TriggerPointSetVar_2","DARTSpectroscopy", 1],
        ["DriveAmplitudeSetVar_3", "DART", 1], 
        ['PV("SampleHeight", {})'.format(value), 4],
    ]
    # Construct the action dict
    for i, key in enumerate(key_list):
        for j in range(len(key)):
            action_dict[key[j]] = value_list[i]

    cmd = action_dict[action]

    if cmd[-1] == 0: # Button
        commands = 'ARExecuteControl("{}","{}",0,"")'.format(cmd[0], cmd[1])
    elif cmd[-1] == 1: # Numeric input
        commands = 'ARExecuteControl("{}","{}",{},"")'.format(cmd[0], cmd[1], value)
    elif cmd[-1] == 2: # String input
        commands = 'ARExecuteControl("{}","{}",0,"{}")'.format(cmd[0], cmd[1], value)
    elif cmd[-1] == 4: # Pure commands
        commands = cmd[0]
    elif cmd[-1] == 5: # Default numeric input
        commands = 'ARExecuteControl("{}","{}",{},"")'.format(cmd[0], cmd[1], cmd[2])
    elif cmd[-1] == 6: # Default string input
        commands = 'ARExecuteControl("{}","{}",0,"{}")'.format(cmd[0], cmd[1], cmd[2])
    else:
        raise ValueError("Function not implemented yet.")

    write_spm(commands=commands, connection=connection, wait=wait, )


# Move the tip 
def move_tip(r, v0=None, s=None, connection=None):
    '''
    Move the tip position based on given displacement.

    Input:
        r   - list: displacement vectors of [x_delta, y_delta]
        v0  - list: voltage applied [V0x, V0y] at the old position (returned from read_spm)
        s   - list: sensitivity of LVDT [sx, sy] for x and y
        connection  - SSH connection: return by return_connection()

    Returns:
        N/A

    Usage:
        keys = ['PIDSLoop.0.Setpoint', 'PIDSLoop.1.Setpoint']
        Vx0, Vy0 = read_spm(key=keys)
        r = [1e-6, 1e-6]
        v0 = [Vx0, Vy0]
        s = [xsens, ysens]
        move_tip(r=r, v0=v0, s=s)
    '''
    x_delta, y_delta = r
    if v0 is None:
        v0x, v0y = read_spm(key=['PIDSLoop.0.Setpoint', 'PIDSLoop.1.Setpoint'], connection=connection)
    else:
        v0x, v0y = v0
    if s is None:
        sx, sy = read_spm(key=['XLVDTSens', 'YLVDTSens'], connection=connection)
    else:
        sx, sy = s
    vx = x_delta / sx
    vy = y_delta / sy

    command1 = 'td_WriteValue("PIDSLoop.0.Setpoint",{})\n'.format(vx+v0x)
    command2 = 'td_WriteValue("PIDSLoop.1.Setpoint",{})\n'.format(vy+v0y)

    write_spm(commands=command1+command2, connection=connection, )


def move_stage(distance, connection=None):
    '''
    Move the stage position based on given displacement.

    Input:
        distance   - list: displacement vectors of [x_delta, y_delta]
        connection - SSH connection: return by utils.return_connection()

    Returns:
        N/A

    Usage:
        move_stage(distance=[x_delta, y_delta])
    '''
    # Determine the moving direction of the stage
    x, y = distance
    x_direction = 'MoveStageLeftButton_0' if x > 0 else 'MoveStageRightButton_0'
    y_direction = 'MoveStageDownButton_0' if y > 0 else 'MoveStageUpButton_0'

    # Change the motor step size for x direction moving
    if x != 0:
        write_spm(commands='PV("StageMoveStepSize", {})'.format(abs(x)), connection=connection)
        write_spm(commands='MoveStage("{}")'.format(x_direction), connection=connection)
    
    if y != 0:
        # Change the motor step size for y direction moving
        write_spm(commands='PV("StageMoveStepSize", {})'.format(abs(y)), connection=connection)
        write_spm(commands='MoveStage("{}")'.format(y_direction), connection=connection)


def tune_probe(num=1, path=os.path.join(_buffer_path, 'Tune.ibw'), center=None, width=50e3, out=False, connection=None):
    '''
    Tune the probe in the DART mode and optionally read the tune data.

    Input:
        num     - Int: how many times tune will be repeated
        path    - String: path where the tune result is saved
        center  - float: center of the tuning frequency range
        width   - float: frequency range of runing
        out     - Boolean: flag to indicate if tune data will be output
    Output:
        tune    - an AR wave containing frequency and amplitude
    Example:
        w = ae.tune_probe(num=2, out=True)
    '''
    for i in range(num):
        spm_control('OneTuneDART', wait=1, connection=connection)
        spm_control('GetTune', wait=0.5, connection=connection)
        if connection is not None:
            aespm.utils.download_file(connection=connection, file_path=path, local_file_name='tune.ibw')
            w = ibw_read('tune.ibw').data
        else:
            w = ibw_read(path).data
        freq = w[0][w[1].argmax()]
        if center is not None:
            if abs(freq-center) > width:
                freq = center
        spm_control('DARTFreq', value=freq, connection=connection)
    spm_control('GetTune', wait=0.5, connection=connection)
    if connection is not None:
        aespm.utils.download_file(connection=connection, file_path=path, local_file_name='tune.ibw')
        w = ibw_read('tune.ibw').data
    else:
        w = ibw_read(path).data
    # Get the freq corresponds to the max intensity
    freq = w[0][w[1].argmax()]
    # Set this freq to be the driven freq
    spm_control('DARTFreq', value=freq, connection=connection)
    spm_control('CenterPhase', connection=connection)
    if out == True:
        return w

def get_files(path, retry=10, sleep_time=6e-3, client=None):
    '''
    Get the file names and sort them in the new to old modification time from a given folder.

    Input:
        path    - String: path to the folder where the SPM data is saved locally
        retry   - Int: number of retries if any of the data file is occupied by SPM controller
        sleep_time - float: wait time between the retries
        client  - SSH client returned by utils.return_connection()
    Output:
        fname   - List: a list of filenames in the order of new to old modification time
    '''
    command = 'dir "{}" /b /o:-d'.format(path)
    retries = 0
    while retries < retry:
        try:
            # Run the command and capture its output
            if client == None:
                result = subprocess.run(command, shell=True, capture_output=True, text=True, check=True)
                file_names = result.stdout.split('\n')
                return [os.path.join(path, file) for file in file_names]
            else:
                stdin, stdout, stderr = client.exec_command(command)
                # Read the output of the command
                file_names = stdout.read().decode('utf-8').splitlines()
                return [os.path.join(path, file) for file in file_names]
            
        except FileNotFoundError:
            print("File not found. Retrying {} times...".format(retries), end='\r')
            retries += 1
            time.sleep(sleep_time)
        except PermissionError:
            print("Permission denied. Retrying {} times...".format(retries), end='\r')
            retries += 1
            time.sleep(sleep_time)
    return 0

def check_file_number(path, wait=1e-1, retry=1e4, client=None):
    '''
    Check if the spectrum/topo measurement is done by monitoring the number of files in the save folder.
    It waits until the number of files has changed to return a True value.
    '''
    num = len(get_files(path, client=client))
    num_new = num
    retries = 0
    # Check the file number in the folder until 16 min is passed
    while num_new == num and retries < retry:
        num_new = len(get_files(path, client=client))
        retries += 1
        time.sleep(wait)
    return True

def ibw_read(fname, retry=10, wait=0.1, lines=False, connection=None):
    '''
    Make a copy of realtime saved image and read the wave data from it.
    Four data channels will be read with the order of ZSenor/Height -> Amplitude -> Phase -> Height/Deflection.
    Each channel contains both the trace and retrace scan lines.
    '''
    retries = 0
    if connection==None:
        if lines==True:
            while retries < retry:
                try:
                    shutil.copy(fname, os.path.join(_buffer_path, 'copy.ibw'))
                    data = bw.load(os.path.join(_buffer_path, 'copy.ibw'))
                    wave = data['wave']['wData']
                    return wave.T
                    # return aespm.tools.load_ibw(fname)
                except FileNotFoundError:
                    print("File not found. Retrying {} times...".format(retries), end='\r')
                    retries += 1
                    time.sleep(wait)
                except PermissionError:
                    print("Permission denied. Retrying {} times...".format(retries), end='\r')
                    retries += 1
                    time.sleep(wait)
        else:
            return aespm.tools.load_ibw(fname)
            # return bw.load(fname)['wave']['wData'].T
    else:
        aespm.utils.download_file(connection=connection, file_path=fname, local_file_name='temp.ibw')
        time.sleep(wait)
        if lines == True:
            return bw.load('temp.ibw')['wave']['wData'].T
        else:
            return aespm.tools.load_ibw('temp.ibw')
        # return bw.load('temp.ibw')['wave']['wData'].T
    return 0

# Function to update and display the plot in real-time
def update_plot_all(folder):
    try:
        while True:
            # Read data from the file
            fname = get_files(folder)
            w = ibw_read(fname[0])

            # Clear the previous plot
            clear_output(wait=True)
            fig,ax = plt.subplots(1,4,figsize=[14,3])
            # Update the plot
            ax[0].plot(w[0], 'r')
            ax[0].plot(w[1], 'b')
            ax[0].set_title("Z Sensor/Height")
            ax[0].set_xlabel("Pixels")
            ax[0].set_ylabel("Z Sensor/Height")

            ax[1].plot(w[2], 'r')
            ax[1].plot(w[3], 'b')
            ax[1].set_title("Amplitude")
            ax[1].set_xlabel("Pixels")
            ax[1].set_ylabel("Amplitude")

            ax[2].plot(w[4], 'r')
            ax[2].plot(w[5], 'b')
            ax[2].set_title("Phase")
            ax[2].set_xlabel("Pixels")
            ax[2].set_ylabel("Phase")

            ax[3].plot(w[6], 'r')
            ax[3].plot(w[7], 'b')
            ax[3].set_title("Height/Deflection")
            ax[3].set_xlabel("Pixels")
            ax[3].set_ylabel("Height/Deflection")

            plt.show()

    except KeyboardInterrupt:
        print("Real-time plotting stopped.")



import paramiko
import Pyro5.api
import subprocess
import os
import platform

import aespm
from aespm import experiment

if platform.system() == 'Windows':
    buffer_path = os.path.join(os.path.expanduser('~'), 'Documents', 'buffer')
    command_buffer = os.path.join(buffer_path, 'ToIgor.arcmd')
    read_out_buffer = os.path.join(buffer_path, 'readout.txt')
    bash_buffer = os.path.join(buffer_path, 'SendToIgor.bat')
    
    path_txt = os.path.join(buffer_path, 'path.txt')
    
    with open(path_txt, 'r') as fopen:
        exe_path = fopen.readline()

class SharedInfo:
    '''
    A class to pass host and local computer directory info across the package.
    Args:
        value (str): IP address of the local computer.
        values (tuple): 
    Returns: None
    '''
    def __init__(self):
        """
        Later come up with better way to initialize the attributes of the class are accessible in multiple scripts.
        """
        self._host = None
        self._buffer_path = None
        self._command_buffer = None
        self._read_out_buffer = None
        self._bash_buffer = None
        self._exe_path = None

        
    def set_host(self, value):
        self._host = value
    
    def set_values(self, values):
        self._buffer_path, self._command_buffer, self._read_out_buffer, self._bash_buffer, self._exe_path = values

    @property
    def host(self):
        return self._host
    
    @property
    
    def buffer_path(self):
        return self._buffer_path
    
    @property
    def command_buffer(self):
        return self._command_buffer
    
    @property
    def read_out_buffer(self):
        return self._read_out_buffer
    
    @property
    def bash_buffer(self):
        return self._bash_buffer
    
    @property
    def exe_path(self):
        return self._exe_path
    
    
shared = SharedInfo()


@Pyro5.api.expose
class CommandExecutor:
    def execute_exe_with_args(self, exe_path, arg):
        try:
            p = subprocess.Popen([exe_path, arg], shell=True)
            p.wait()
        except Exception as e:
            return f"Execution Failed: {str(e)}"
    def get_local_dir(self):
        return "$".join([buffer_path, command_buffer, read_out_buffer, bash_buffer, exe_path])

def connect(host, port=9091):
    '''
    Start a python server on your local computer. This server can accept
    commands from a remote cluster.

    Args:
        host    - String: IP address of the local computer.
        port    - Int (Optional): the port for the server.
        
    Output: None
    Example:
        >>> import aespm as ae
        >>> host = 'your_ip_address'
        >>> ae.utils.connect(host=host)
    '''
    daemon = Pyro5.server.Daemon(host=host, port=port)  # Bind to the specific IP
    uri = daemon.register(CommandExecutor, "command.executor")
    # print("Server is ready. Object uri =", uri)
    daemon.requestLoop()  # Start the request loop

def get_local_directory(host=shared._host, port=9091):
    """Retrieve the buffer path on the local computer.

    Args:
        uri (string): address of server
        args (string): arguments to pass to the executable
    
    Returns: "$".join([buffer_path, command_buffer, read_out_buffer, bash_buffer, exe_path])
    """
    
    uri = "PYRO:command.executor@{}:{}".format(host, port)
    command_executor = Pyro5.api.Proxy(uri)
    response = command_executor.get_local_dir()
    return response

def execute_exe_on_server(exe_path, args, host=shared._host, port=9091):
    """Execute an executable on the remote server.

    Args:
        exe_path (string): path to the executable
        uri (string): address of server
        args (string): arguments to pass to the executable
    
    Returns: None
    """
    uri = "PYRO:command.executor@{}:{}".format(host, port)
    #   print("Connecting to server at", uri)
    command_executor = Pyro5.api.Proxy(uri)
    response = command_executor.execute_exe_with_args(exe_path, args)
    #   print("Server response:", response)

def main_exe_on_server(host=shared._host):
    """Execute the main executable on the remote server.
    """
    exe_path = shared._exe_path
    # args = aespm._command_buffer
    execute_exe_on_server(exe_path, args=shared._command_buffer, host=host)

def return_connection(host: str, username: str, password: str):
    """
    Return a connection object to the remote server.
    Args:
        host (str): IP address of the remote server.
        username (str): Username for the remote server.
        password (str): Password for the remote server.
    Returns:
        connection (object): Connection object to the remote server.
    """
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    client.connect(host, username=username, password=password)
    connection = client.open_sftp()
    return connection, client

def read_remote_file(connection: object, file_path: str):
    """
    Read content from the remote file.
    Args:
        connection (object): Connection object to the remote server.
        file_path (str): Path to the remote file.
    Returns:
        content (str): Content of the remote file.
    """
    with connection.open(file_path, 'r') as remote_file:
        content = remote_file.read()
    return content

def write_to_remote_file(connection: object, file_path: str, data: str):
    """
    Write data to the remote file.
    Args:
        connection (object): Connection object to the remote server.
        file_path (str): Path to the remote file.
        data (str): Data to be written to the remote file.
    Returns:
        None
    """
    with connection.open(file_path, 'w') as remote_file:
        remote_file.write(data)

def download_file(connection: object, file_path: str, local_file_name: str = "isaac_copy.ibw"):
    """
    Download a file from the remote server to the local machine.
    Args:
        connection (object): Connection object to the remote server.
        file_path (str): Path to the remote file.
        local_file_name (str): Name of the local file.
    Returns:
        None
    """
    connection.get(file_path, local_file_name)




import warnings
warnings.filterwarnings("ignore", category=SyntaxWarning)
#For Python 3.8+, to avoid SyntaxWarning: "is" with a literal. Did you mean "=="?

__all__ = ['experiment']

# If there is no buffer files, create them in Documents/buffers
import os 
import shutil
import platform

doc_path = os.path.expanduser('~')
buffer_path = os.path.join(doc_path, 'Documents', 'buffer')
    
# Initialize the buffer files only on Windows operating system

if platform.system() == 'Windows':

    # Load the path from file if there is a path.txt in the buffer folder
    path_txt = os.path.join(buffer_path, 'path.txt')
    
    command_buffer = os.path.join(buffer_path, 'ToIgor.arcmd')
    read_out_buffer = os.path.join(buffer_path, 'readout.txt')
    bash_buffer = os.path.join(buffer_path, 'SendToIgor.bat')
    
    try:
        # Create the buffer folder
        if not os.path.exists(buffer_path):
            os.makedirs(buffer_path)

        # Create the commands buffer file
        if not os.path.exists(command_buffer):
            with open(command_buffer, 'w') as fopen:
                fopen.write(' ')
                
        if os.path.exists(path_txt):
            # Load the exe path from the file
            with open(path_txt, 'r') as fopen:
                exe_path = fopen.readline()
            
            # Write the exe_path into the bash buffer
            with open(bash_buffer, 'w') as fopen:
                fopen.write('"{}" "{}"'.format(exe_path, command_buffer))
        else:
            # Fist-time installation: no path_txt file available
            # Detect the AR version and installation path automatically
            if os.path.exists("C:\\AsylumResearch\\v19"):
                exe_path = "C:\\AsylumResearch\\v19\\RealTime\\Igor Pro Folder\\Igor.exe"
            elif os.path.exists("C:\\AsylumResearch\\v18"):
                exe_path = "C:\\AsylumResearch\\v18\\RealTime\\Igor Pro Folder\\Igor.exe"
            elif os.path.exists("C:\\AsylumResearch\\v17"):
                exe_path = "C:\\AsylumResearch\\v17\\RealTime\\Igor Pro Folder\\Igor.exe"
            elif os.path.exists("C:\\AsylumResearch\\v16"):
                exe_path = "C:\\AsylumResearch\\v16\\RealTime\\Igor Pro Folder\\Igor.exe"
            elif os.path.exists("C:\\AsylumResearch\\v15"):
                exe_path = "C:\\AsylumResearch\\v15\\RealTime\\Igor Pro Folder\\Igor.exe"
            elif os.path.exists("C:\\AsylumResearch\\v14"):
                exe_path = "C:\\AsylumResearch\\v14\\RealTime\\Igor Pro Folder\\Igor.exe"
            elif os.path.exists("C:\\AsylumResearch\\v13"):
                exe_path = "C:\\AsylumResearch\\v13\\RealTime\\Igor Pro Folder\\Igor.exe"
            else:
                print("No AR detected. Please modify path.txt manually to include the AR path.")
        
            with open(path_txt, 'w') as fopen:
                fopen.write(exe_path)
                
            with open(bash_buffer, 'w') as fopen: 
                fopen.write('"{}" "{}"'.format(exe_path, command_buffer))

            # Copy the user functions into the default include folder of AR
            # shutil.copy(os.path.join(aespm.__path__, 'user functions', 'UserFunctions.ipf'), os.path.join(doc_path, 'AsylumResearch', 'UserIncludes'))

    except PermissionError:
        print('No writing permission to ~/Documents. Please create buffer folder and files manually.')
        
from aespm.experiment import *

from aespm import tools
from aespm import utils


from setuptools import setup, find_packages

setup(
  name = 'aespm',
  version = '1.1.0',
  packages = find_packages(),
  license='MIT',
  description = 'SPM Automation with Python and Machine Learning.',
  long_description = 'Python interface that enables local and remote control of Scanning Probe Microscope (SPM) with codes.\nIt offers a modular way to write autonomous workflows ranging from simple routine operations, to advanced automatic scientific discovery based on machine learning.',
  author = ["Richard (Yu) Liu"],
  email = ['yu93liu@gmail.com'],
  #authors = [
  #    { name="Richard (Yu) Liu", email="yu93liu@gmail.com" },
  #    { name="Boris Slautin", email="bslautin@gmail.com" },
  #],
  url = 'https://github.com/RichardLiuCoding/aespm',
  download_url = 'https://github.com/RichardLiuCoding/aespm.git',
  keywords = ['SPM', 'Python', 'Instrument control', 'Autonomoous', 'Machine learning','Data Analysis'],
  install_requires=[
          'numpy',
          'scipy',
          'matplotlib',
          'igor2',
          'paramiko',
          'Pyro5',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3'
  ],
)
